#include "../config.h"
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <errno.h>
#include <sys/poll.h>
#include "vcr_compress.h"

#include "nesvideos-piece.hh"
#include <cmath>
#include "filters.h"
#include <vector>
#include <algorithm>

void
VCRComp_init(){}

static long wid;
static long hei;
static int fp;

static int audiolead;
static int audioperframe;
static std::vector<unsigned char> lastframe;

#define MagickPI  3.14159265358979323846264338327950288419716939937510
static inline double Sinc(double x)
{
  x *= MagickPI;
  return std::sin(x) / x;
}
static inline double Lanczos(double x)
{
  if(x < 0) x = -x;
  if(x == 0.0) return 1.0;
  if(x >= 3.0)return 0.0;
  return Sinc(x) * Sinc(x/3.0);
}

template<typename InIt, typename OutIt>
static inline void Resample
    (const FilterInfo&filter,
     InIt & source,  unsigned src_len,
     OutIt& result, unsigned dest_len)
{
    const double blur = 1.0;
    const double factor = dest_len / (double)src_len;
    
    double contribution[dest_len];
    
    double scale   = std::min(factor, 1.0) / blur;
    double support = filter.support / scale;
    if(support <= 0.5) { support = 0.5 + 1E-12; scale = 1.0; }
    
    for(unsigned x=0; x<dest_len; ++x)
    {
        double center = (x+0.5) / factor;
        unsigned start = (unsigned)std::max(center-support+0.5, (double)0);
        unsigned stop  = (unsigned)std::min(center+support+0.5, (double)src_len);
        double density = 0.0;
        unsigned nmax = stop-start;
        double s = start - center+0.5;
        for(unsigned n=0; n<nmax; ++n, ++s)
        {
            contribution[n] =
                Lanczos/*filter.function*/
                (s * scale);
            density += contribution[n];
        }
        if(density != 0.0 && density != 1.0)
        {
            for(unsigned n=0; n<nmax; ++n)
                contribution[n] /= density;
        }
        result.Zero(x);
        for(unsigned n=0; n<nmax; ++n)
        {
            result.Add(x, source, start+n, contribution[n]);
        }
        result.Lock(x);
    }
}

struct Proxy1 // temporary
{
    inline Proxy1(unsigned _W) : W(_W)
    {
        data.resize(hei * W * 3);
    }
    inline void Zero(unsigned pos)
    {
        /*
        for(unsigned y=0; y<hei; ++y)
            for(unsigned c=0; c<3; ++c)
                data[y*W*3 + pos*3+c] = 0;
        */
    }
    inline void Add(unsigned pos, const unsigned char*p, unsigned ppos, double value)
    {
        unsigned srcpos  = ppos*3;
        unsigned destpos = pos*3;
        for(unsigned y=0; y<hei; ++y)
        {
            data[destpos+0] += p[srcpos+0] * value;
            data[destpos+1] += p[srcpos+1] * value;
            data[destpos+2] += p[srcpos+2] * value;
            srcpos  += wid*3;
            destpos +=   W*3;
        }
    }
    inline void Lock(unsigned pos)
    {
    }
public:
    std::vector<double> data;
    unsigned W;
};
struct Proxy2 // final
{
    inline Proxy2(unsigned _W,unsigned _H) : W(_W)
    {
        data.resize(W*_H*3);
        result.resize(W*_H*3);
    }
    inline void Zero(unsigned pos)
    {
        /*
        for(unsigned x=0; x<W*3; ++x)
            data[pos*W*3 + x] = 0;
        */
    }
    inline void Add(unsigned pos, Proxy1& p, unsigned ppos, double value)
    {
        unsigned destpos = pos*W*3;
        unsigned srcpos  = ppos*W*3;
        for(unsigned x=0; x<W*3; ++x)
            data[destpos + x] += p.data[srcpos+x]*value;
    }
    inline void Lock(unsigned pos)
    {
        unsigned srcpos = pos*W*3;
        for(unsigned x=0; x<W*3; ++x)
        {
            int col = (int)(data[srcpos + x]);
            result[srcpos + x] = std::max(std::min(col, 255), 0);
        }
    }
public:
    std::vector<double> data;
    std::vector<unsigned char> result;
    unsigned W;
};

static void Put(const unsigned char* data, bool Cached)
{
    const unsigned W = 512;
    const unsigned H = 384;
    
    if(W == wid && H == hei)
    {
        if(!Cached)
            lastframe.assign(data, data + wid*hei*3);
        Cached=true;
    }
    
	//NESVideoLoggingVideo(data, wid,hei,fp<<24);
	if(!Cached)
	{
	    #define FILTER_NONE       0
	    #define FILTER_LINEAR     1
		#define FILTER_TRIANGLE   2
		#define FILTER_BLACKMAN   6
		#define FILTER_QUADRATIC  8
		#define FILTER_CUBIC      9
		#define FILTER_CATROM     10
		#define FILTER_LANCZOS    12
		#define FILTER_BESSEL     13
		#define FILTER_SINC       14
		
		Proxy1 proxy1(W);
		
		Resample(Filters[FILTER_LANCZOS], data,wid, proxy1,W);
		
		Proxy2 proxy2(W,H);
		
		Resample(Filters[FILTER_LANCZOS], proxy1,hei, proxy2,H);
		
		NESVideoLoggingVideo(&proxy2.result[0], W,H, fp<<24);
		
		lastframe.swap(proxy2.result);
	}
	else
	{
		NESVideoLoggingVideo(data, W,H, fp<<24);
	}
}

static void VideoAdd(const unsigned char *data, int size, bool Cached=false)
{
	if (size != wid*hei*3)
	{
		fprintf(stderr, "VideoAdd: %d*%d*3 != %d \n", wid, hei, size);
		abort();
	}
	
	if(LoggingEnabled < 2) return;
	
    // If the audio is lagging too far behind, have mercy
    if (audiolead < (-3*audioperframe))
        return;

	Put(data, Cached);
	
	audiolead -= audioperframe;
}

static void AudioAdd(const unsigned char *data, int size)
{
    if(LoggingEnabled < 2) return;
    
    NESVideoLoggingAudio(data, 44100, 16,2,
      size/2/2
                        );
    
    audiolead += size/2/2;
    while(audiolead > (3*audioperframe))
    {
        VideoAdd(&lastframe[0], lastframe.size(), true);
    }
}

void
VCRComp_startFile( const char *filename, long width, long height, int fps )
{
    NESVideoSetVideoCmd(filename);
    
    wid=width;
    hei=height;
    fp=fps;
    
    LoggingEnabled=2;

	audiolead = 0;
	audioperframe = 44100 / fp;
}

void
VCRComp_finishFile()
{
    LoggingEnabled=0;
	return;
}

void VCRComp_addVideoFrame( const unsigned char *data )
{
	VideoAdd(data, wid*hei*3);
}

void
VCRComp_addAudioData( const unsigned char *data, int len )
{
	AudioAdd(data,len);
}

int
VCRComp_numVideoCodecs()
{
	return 1;
}

const char *
VCRComp_videoCodecName( int index )
{
	return "Raw Video (BGR or BGR24)";
}

int
VCRComp_numVideoCodecAttribs( int index )
{
	return 1;
}

const char *
VCRComp_videoCodecAttribName( int cindex, int aindex )
{
	return "<No settings>";
}

int
VCRComp_videoCodecAttribKind( int cindex, int aindex )
{
	return ATTRIB_STRING;
}

const char *
VCRComp_videoCodecAttribValue( int cindex, int aindex )
{
	return "-1";
}

void
VCRComp_videoCodecAttribSetValue( int cindex, int aindex, const char *val )
{

}

int
VCRComp_numVideoCodecAttribOptions( int cindex, int aindex )
{
	return 0;
}

const char *
VCRComp_videoCodecAttribOption( int cindex, int aindex, int oindex )
{
	return "--";
}

void
VCRComp_selectVideoCodec( int index )
{}

int
VCRComp_numAudioCodecs()
{
	return 1;
}

const char *
VCRComp_audioCodecName( int index )
{
	return "Raw Audio";
}

int
VCRComp_numAudioCodecAttribs( int index )
{
	return 1;
}

const char *
VCRComp_audioCodecAttribName( int cindex, int aindex )
{
	return "<No settings>";
}

int
VCRComp_audioCodecAttribKind( int cindex, int aindex )
{
	return ATTRIB_INTEGER;
}

const char *
VCRComp_audioCodecAttribValue( int cindex, int aindex )
{
	return "-1";
}

void
VCRComp_audioCodecAttribSetValue( int cindex, int aindex, const char *val )
{}


int
VCRComp_numAudioCodecAttribOptions( int cindex, int aindex )
{
	return 0;
}

const char *
VCRComp_audioCodecAttribOption( int cindex, int aindex, int oindex )
{
	return "--";
}

void
VCRComp_selectAudioCodec( int index )
{}

